/*!
 * \file comm_spi_if.c
 * \brief CPU間通信SPI ドライバ
 * \author Hisashi IIDA
 *
 * \note
 * Copyright(C) 2013 Yamaha Corporation<br>
 * All rights reserved.<br>
 * $Id:$
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/uaccess.h>  /* for copy_from_user / copy_to_user */

#include <linux/fs.h>       /* for struct file_operations */
#include <linux/cdev.h>
#include <linux/slab.h>
#include <linux/spi/spi.h>

#include "comm_spi_if.h"

#define MAX_REG_MINOR (1)

#define COMM_SPI_CTX_COMM_SPI   (0)
#define COMM_SPI_CTX_MAX        (1)

#define SPI_WORD_SIZE_8_BIT     (8)
#define SPI_WORD_SIZE_16_BIT    (16)
#define SPI_WORD_SIZE_32_BIT    (32)
#define SPI_WORD_SIZE_TRANSFER       SPI_WORD_SIZE_32_BIT 
#define SPI_WORD_SIZE_LOOPBACK       SPI_WORD_SIZE_8_BIT

#define COMM_SPI_DBG_LOG(x, ...)    \
    do {  \
        printk("[%s, %d]: ", __FILE__, __LINE__);   \
        printk(x, ## __VA_ARGS__); \
        printk("\n");   \
    } while(0)

#define ROUNDUP(_a, _s)     ((((_a) + (_s) - 1) / (_s)) * (_s))

/* class_create / device_create を使うためには下記の宣言が必要 */
MODULE_LICENSE("GPL");

/* ドライバコンテキスト */
struct __YAMAHA_CommSpiCtx_struct {
    struct cdev                     cdev;
    dev_t                           dev;
    struct class                    *class;
    const struct file_operations    *fops;
    struct spi_device               *spi_device;
    struct spi_driver               *spi_driver;
    const char                      *name;

    char                            *txbuf;
    char                            *rxbuf;
    unsigned long                   len;
};

typedef struct __YAMAHA_CommSpiCtx_struct CommSpiCtx;

static long comm_spi_ioctl_comm_spi(struct file *file,
                                   unsigned int cmd, unsigned long a);
static int comm_spi_ioctl(CommSpiCtx *ctx,
                                unsigned int cmd, unsigned long a);
static int comm_spi_ioctl_transfer(CommSpiCtx *ctx, unsigned long a);
static void comm_spi_transfer_setup_transfer(struct spi_transfer *transfer,
                    CommSpiCtx *ctx, commSpiIoctlTransferParam *p);
static int comm_spi_byte_swap(unsigned char *buf, unsigned int bufsize,
                                unsigned int wordsize);
static int comm_spi_ioctl_loopback(CommSpiCtx *ctx, unsigned long a);
static void comm_spi_loopback_setup_transfer(struct spi_transfer *transfer,
                    CommSpiCtx *ctx, commSpiIoctlLoopbackParam *p);
static int comm_spi_init(void);
static void comm_spi_exit(void);
static int comm_spi_probe(CommSpiCtx *ctx);
static int comm_spi_probe_comm_spi(struct spi_device *spi);
static int comm_spi_remove_comm_spi(struct spi_device *spi);

/* ファイルオペレーション構造体 */
static const struct file_operations CommSpiFops[]= {
    {
        .owner      = THIS_MODULE,
        .unlocked_ioctl      = comm_spi_ioctl_comm_spi,
    }
};

static const char *CommSpiName[] = {
    COMM_SPI_TRANSFER_DEVICE_NAME,
};

static CommSpiCtx __CommSpiCtx[COMM_SPI_CTX_MAX];

/* SPIドライバ登録情報 */
static struct spi_driver CommSpiDriver[] = {
    {
        .driver = {
            .name   = COMM_SPI_TRANSFER_DEVICE_NAME,
            .bus    = &spi_bus_type,
            .owner  = THIS_MODULE,
        },
        .probe      = comm_spi_probe_comm_spi,
        .remove     = comm_spi_remove_comm_spi,
    },
};

/*!
 * \brief ドライバのioctlハンドラ
 * \param cmd[IN]: ioctlコマンド
 *        a[IN/OUT]: ユーザの通信バッファ
 *
 * \return 0 : 成功
 *         負：失敗
 */
static long
comm_spi_ioctl_comm_spi(struct file *file,
                                   unsigned int cmd, unsigned long a)
{
    CommSpiCtx *ctx;

    ctx = &__CommSpiCtx[COMM_SPI_CTX_COMM_SPI];

    return (comm_spi_ioctl(ctx, cmd, a));
}

/*!
 * \brief ドライバのioctlハンドラ
 * \param ctx[IN]: ドライバコンテキスト
 *        cmd[IN]: ioctlコマンド
 *        a[IN/OUT]: ユーザの通信バッファ
 *
 * \return 0 : 成功
 *         負：失敗
 */
static int
comm_spi_ioctl(CommSpiCtx *ctx, unsigned int cmd, unsigned long a)
{
    int rc;

    switch (cmd) {
    case COMM_SPI_IOCTL_TRANSFER_CMD:
    {
        rc = comm_spi_ioctl_transfer(ctx, a);
        if (rc != 0) {
            return (rc);
        }

        break;
    }
    case COMM_SPI_IOCTL_LOOPBACK_CMD:
    {
        rc = comm_spi_ioctl_loopback(ctx, a);
        if (rc != 0) {
            return (rc);
        }

        break;
    }
    case COMM_SPI_IOCTL_SET_DMA_FLAG_CMD:
    {
        commSpiIoctlDMAFlag flag;

        rc = copy_from_user(&flag, (void*)a, sizeof(flag));
        if (rc != 0) {
            COMM_SPI_DBG_LOG("%s: copy_from_user() failed", ctx->name);
            return (rc);
        }

        spi_sync_set_DMA_started_flag(flag);

        break;
    }
    case COMM_SPI_IOCTL_GET_DMA_FLAG_CMD:
    {
        commSpiIoctlDMAFlag flag;

        flag = spi_sync_get_DMA_started_flag();

        rc = copy_to_user((commSpiIoctlDMAFlag*)a, &flag, sizeof(flag));
        if (rc != 0) {
            COMM_SPI_DBG_LOG("%s: copy_to_user() failed", ctx->name);
            return (rc);
        }

        break;
    }
    default:
        COMM_SPI_DBG_LOG("Invalid param: cmd= %u", cmd);
        return (-EINVAL);
    }

    return (0);
}

/*!
 * \brief transferの処理
 * \param ctx[IN]: コンテキスト
 *        a[IN/OUT]: ユーザの通信バッファ
 *
 * \return 0 : 成功
 *         負：失敗
 */
static int
comm_spi_ioctl_transfer(CommSpiCtx *ctx, unsigned long a)
{
    commSpiIoctlTransferParam p;
    struct spi_message  message;
    struct spi_transfer transfer;
    int rc;
    unsigned int wordsize; /* word size (in byte unit) */ 

    wordsize = SPI_WORD_SIZE_TRANSFER  / 8;

    /* Set user's data into commSpiIoctlTransferParam structure */
    rc = copy_from_user(&p, (void*)a, sizeof(p));
    if (rc != 0) {
        COMM_SPI_DBG_LOG("%s copy_from_user() failed", ctx->name);
        return (rc);
    }

    /* Round up user buffer's length to the next multiple of word size */
    ctx->len = ROUNDUP(p.bufsize, wordsize);
    if (ctx->len > COMM_SPI_BUF_SIZE) {
        COMM_SPI_DBG_LOG("Invalid param: bufsize = %lu", p.bufsize);
        return (-EINVAL);
    }

    /* Copy user's buffer contents to context buffer */
    if (p.txbuf) {
        memset(ctx->txbuf, 0xFF, ctx->len);
        rc = copy_from_user(ctx->txbuf, p.txbuf, p.bufsize);
        if (rc != 0) {
            COMM_SPI_DBG_LOG("%s copy_from_user() failed", ctx->name);
            return (rc);
        }
 
        /* At the SPI master's end, the data will arrive reversed at every word size. */
        /* Example: For 4byte wordsize, data sent "TRUE" will arrive "EURT" */
        /* Example: For 2byte wordsize, data sent "US" will arrive "SU" */
        /* Example: For 1byte wordsize, data sent "A" will arrive "A" (unchanged) */
        /* To avoid the labor at SPI master's end, of reversing the data back to normal, */
        /* here we purposely arrange them backwards. */
        rc = comm_spi_byte_swap(ctx->txbuf, ctx->len, wordsize);
        if (rc != 0) {
            COMM_SPI_DBG_LOG("comm_spi_byte_swap() failed");
            return (rc);
        }
    }

    /* Set spi_transfer */
    comm_spi_transfer_setup_transfer(&transfer, ctx, &p);

    /* Execute transfer */
    spi_message_init(&message);
    spi_message_add_tail(&transfer, &message);
    rc = spi_sync(ctx->spi_device, &message);
    if (rc != 0) {
        return (rc);
    }

    /* Copy context's buffer content to user's buffer */
    /* Received data will arrive reversed at every word size. */
    /* So, reverse them to the original order */
    if (p.rxbuf) {
        rc = comm_spi_byte_swap(ctx->rxbuf, ctx->len, wordsize);
        if (rc != 0) {
            COMM_SPI_DBG_LOG("comm_spi_byte_swap() failed");
            return (rc);
        }

        rc = copy_to_user(p.rxbuf, ctx->rxbuf, p.bufsize);
        if (rc != 0) {
            COMM_SPI_DBG_LOG("%s copy_from_user() failed", ctx->name);
            return (rc);
        }
    }

    return (0);
}

/*!
 * \brief 通信情報構造体のセットアップ
 * \param transfer[OUT]: 通信情報構造体
 *        ctx[IN]: コンテキスト
 *        p[IN]: ユーザからの通信情報
 *
 * \return 0 : 成功
 *         負：失敗
 */
static void
comm_spi_transfer_setup_transfer(struct spi_transfer *transfer,
                    CommSpiCtx *ctx, commSpiIoctlTransferParam *p)
{
    memset(transfer, 0, sizeof(*transfer));

    if (p->txbuf) {
        transfer->tx_buf = ctx->txbuf;
    }
    if (p->rxbuf) {
        transfer->rx_buf = ctx->rxbuf;
    }
    transfer->len = ctx->len;

    if (p->speed_hz != 0) {
        transfer->speed_hz = p->speed_hz;
    }
    transfer->bits_per_word = SPI_WORD_SIZE_TRANSFER;
}

/*!
 * \brief loopbackの処理
 * \param ctx[IN]: コンテキスト
 *        a[IN/OUT]: ユーザの通信バッファ
 *
 * \return 0 : 成功
 *         負：失敗
 */
static int
comm_spi_ioctl_loopback(CommSpiCtx *ctx, unsigned long a)
{
    commSpiIoctlLoopbackParam p;
    struct spi_message  message;
    struct spi_transfer transfer;
    int rc;
    unsigned int wordsize; /* word size (in byte unit) */ 

    wordsize = SPI_WORD_SIZE_LOOPBACK / 8;

    /* Set user's data into commSpiIoctlLoopbackParam structure */
    rc = copy_from_user(&p, (void*)a, sizeof(p));
    if (rc != 0) {
        COMM_SPI_DBG_LOG("%s copy_from_user() failed", ctx->name);
        return (rc);
    }

    /* Round up user buffer's length to the next multiple of word size */
    ctx->len = ROUNDUP(p.bufsize, wordsize);
    if (ctx->len > COMM_SPI_BUF_SIZE) {
        COMM_SPI_DBG_LOG("Invalid param: bufsize = %lu", p.bufsize);
        return (-EINVAL);
    }

    /* Copy user's buffer contents to context buffer */
    if (!p.txbuf) {
        COMM_SPI_DBG_LOG("NULL txbuf");
        return (-EINVAL);
    }
    if (!p.rxbuf) {
        COMM_SPI_DBG_LOG("NULL rxbuf");
        return (-EINVAL);
    }
    rc = copy_from_user(ctx->txbuf, p.txbuf, p.bufsize);
    if (rc != 0) {
        COMM_SPI_DBG_LOG("%s copy_from_user() failed", ctx->name);
        return (rc);
    }
    /* At the SPI master's end, the data will arrive at reversed every word size. */
    /* Example: For 4byte wordsize, data sent "TRUE" will arrive "EURT" */
    /* Example: For 2byte wordsize, data sent "US" will arrive "SU" */
    /* Example: For 1byte wordsize, data sent "A" will arrive "A" (unchanged) */
    /* To avoid the labor of reversing the data back to normal, */
    /* here we purposely arrange them backwards. */
    rc = comm_spi_byte_swap(ctx->txbuf, ctx->len, wordsize);
    if (rc != 0) {
        COMM_SPI_DBG_LOG("comm_spi_byte_swap() failed");
        return (rc);
    }

    /* Set spi_transfer */
    comm_spi_loopback_setup_transfer(&transfer, ctx, &p);

    /* Execute transfer */
    spi_message_init(&message);
    spi_message_add_tail(&transfer, &message);
    rc = spi_sync(ctx->spi_device, &message);
    if (rc != 0) {
        return (rc);
    }

    /* Copy context's buffer content to user's buffer */
    /* Received data will arrive reversed at every word size. */
    /* So, reverse them to the original order */
    rc = comm_spi_byte_swap(ctx->rxbuf, ctx->len, wordsize);
    if (rc != 0) {
        COMM_SPI_DBG_LOG("comm_spi_byte_swap() failed");
        return (rc);
    }
    rc = copy_to_user(p.rxbuf, ctx->rxbuf, p.bufsize);
    if (rc != 0) {
        COMM_SPI_DBG_LOG("%s copy_from_user() failed", ctx->name);
        return (rc);
    }

    return (0);
}

/*!
 * \brief 通信情報構造体のセットアップ
 * \param transfer[OUT]: 通信情報構造体
 *        ctx[IN]: コンテキスト
 *        p[IN]: ユーザからの通信情報
 *
 * \return 0 : 成功
 *         負：失敗
 */
static void
comm_spi_loopback_setup_transfer(struct spi_transfer *transfer,
                    CommSpiCtx *ctx, commSpiIoctlLoopbackParam *p)
{
    memset(transfer, 0, sizeof(transfer));

    transfer->tx_buf = ctx->txbuf;
    transfer->rx_buf = ctx->rxbuf;
    transfer->len = ctx->len;

    if (p->speed_hz != 0) {
        transfer->speed_hz = p->speed_hz;
    }
    transfer->bits_per_word = SPI_WORD_SIZE_LOOPBACK;
}

/*!
 * \brief 送信する前にTxバッファーのデータの並べ替え処理
 * \param ctx[OUT] : Txバッファ
 *        bufsize[OUT] : Txバッファ長 （必ず4の倍数を指定すること）
 *
 * \return 0 : 成功
 *         負：失敗
 */
static int
comm_spi_byte_swap(unsigned char *buf, unsigned int bufsize, unsigned int wordsize)
{
    unsigned char tmpbuf[4];
    unsigned char *sp;
    unsigned int i, rep;

    switch (wordsize) {
    case 1:
        break;
    case 2:
        rep = bufsize / 2; // Loop repetitions. Every loop handles 2 bytes of buffer 
        sp = buf;
 
        for (i=0; i < rep; i++) {
            tmpbuf[0] = sp[0];
            tmpbuf[1] = sp[1];
 
            sp[0] = tmpbuf[1];
            sp[1] = tmpbuf[0];
 
            sp += 2;
        }
        break;
    case 4:
        rep = bufsize / 4; // Loop repetitions. Every loop handles 4 bytes of buffer 
        sp = buf;
 
        for (i=0; i < rep; i++) {
            tmpbuf[0] = sp[0];
            tmpbuf[1] = sp[1];
            tmpbuf[2] = sp[2];
            tmpbuf[3] = sp[3];
 
            sp[0] = tmpbuf[3];
            sp[1] = tmpbuf[2];
            sp[2] = tmpbuf[1];
            sp[3] = tmpbuf[0];
 
            sp += 4;
        }
        break;
    default:
        COMM_SPI_DBG_LOG("Invalid wordsize %u", wordsize);
        return (-EINVAL);
    }

    return (0);
}

/*!
 * \brief ドライバのロード処理
 *
 * \return 0 : 成功
 */
static int
comm_spi_init(void)
{
    CommSpiCtx *ctx;
    int rc, i;

    for (i = 0; i < ARRAY_SIZE(__CommSpiCtx); i++) {
        ctx = &__CommSpiCtx[i];

        ctx->name = CommSpiName[i];
        ctx->fops = &CommSpiFops[i];

        ctx->spi_driver = &CommSpiDriver[i];
        ctx->spi_device = NULL;

        /* SPIドライバの登録 */
        rc = spi_register_driver(ctx->spi_driver);
        if (rc != 0) {
            COMM_SPI_DBG_LOG("spi_register_driver() = %d", rc);
            unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
            device_destroy(ctx->class, ctx->dev);
            class_destroy(ctx->class);
            cdev_del(&ctx->cdev);
            return (rc);
        }
    }

    return (0);
}

/*!
 * \brief ドライバのprobeハンドラ
 * \param ctx[OUT]: ドライバコンテキスト
 *
 * \return 0 : 成功
 *         負：失敗
 */
static int
comm_spi_probe(CommSpiCtx *ctx)
{
    dev_t devno;
    int rc;

    /* メジャー番号の動的な取得 */
    rc = alloc_chrdev_region(&devno, 0, MAX_REG_MINOR, ctx->name);
    if (rc != 0) {
        COMM_SPI_DBG_LOG("alloc_chrdev_region(%s) failed", ctx->name);
        return (-ENOMEM);
    }

    /* キャラクタ・デバイスのデバイス番号の登録 */
    ctx->dev = MKDEV(MAJOR(devno), 0);

    cdev_init(&ctx->cdev, ctx->fops);
    rc = cdev_add(&ctx->cdev, ctx->dev, MAX_REG_MINOR);
    if (rc != 0) {
        COMM_SPI_DBG_LOG("cdev_add() = %d", rc);
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        return (-ENOMEM);
    }

    /* /dev/xxx のデバイスノードの作成 */
    ctx->class = class_create(THIS_MODULE, ctx->name);
    if (IS_ERR(ctx->class)) {
        COMM_SPI_DBG_LOG("class_create() failed");
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        cdev_del(&ctx->cdev);
        return (PTR_ERR(ctx->class));
    }

    device_create(ctx->class, NULL, ctx->dev, NULL, ctx->name);

    return (0);
}

/*!
 * \brief ドライバのアンロード処理
 *
 * \return 0 : 成功
 */
static void
comm_spi_exit(void)
{
    CommSpiCtx *ctx;
    int i;

    for (i = 0; i < ARRAY_SIZE(__CommSpiCtx); i++) {
        ctx = &__CommSpiCtx[i];

        spi_unregister_driver(ctx->spi_driver);
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        device_destroy(ctx->class, ctx->dev);
        class_destroy(ctx->class);
        cdev_del(&ctx->cdev);
    }

    return;
}

/*!
 * \brief ドライバのprobeハンドラ
 * \param spi[IN]: SPIデバイス
 *
 * \return 0 : 成功
 *         負：失敗
 */
static int
comm_spi_probe_comm_spi(struct spi_device *spi)
{
    CommSpiCtx *ctx;

    ctx = &__CommSpiCtx[COMM_SPI_CTX_COMM_SPI];

    ctx->spi_device = spi;
    ctx->txbuf = kmalloc(COMM_SPI_BUF_SIZE, GFP_KERNEL);
    ctx->rxbuf = kmalloc(COMM_SPI_BUF_SIZE, GFP_KERNEL);

    return (comm_spi_probe(ctx));
}

/*!
 * \brief ドライバのremoveハンドラ
 *
 * \return 0 : 成功
 *         負：失敗
 */
static int
comm_spi_remove_comm_spi(struct spi_device *spi)
{
    CommSpiCtx *ctx;

    ctx = &__CommSpiCtx[COMM_SPI_CTX_COMM_SPI];

    (void)spi;
    kfree(ctx->txbuf);
    kfree(ctx->rxbuf);
    ctx->spi_device = NULL;

    return (0);
}

module_init(comm_spi_init);
module_exit(comm_spi_exit);
