/*!
 * \file gpio_ctrl_if.c
 * \brief GPIO ドライバ
 * \author Hisashi IIDA
 *
 * \note
 * Copyright(C) 2014 Yamaha Corporation<br>
 * All rights reserved.<br>
 * $Id:$
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/uaccess.h>  /* for copy_from_user / copy_to_user */

#include <linux/fs.h>       /* for struct file_operations */
#include <linux/cdev.h>
#include <linux/slab.h>

#include <linux/gpio.h>

#include "gpio_ctrl.h"

/* class_create / device_create を使うためには下記の宣言が必要 */
MODULE_LICENSE("GPL");

static long gpio_ioctl(struct file *file, unsigned int command, unsigned long arg);
static void gpio_ctrl_exit(void);
static int gpio_ctrl_init(void);

/* ファイルオペレーション構造体 */
static const struct file_operations gpio_ctrl_fops = {
    .owner          = THIS_MODULE,
    .unlocked_ioctl = gpio_ioctl,
};

/* Convert GPIO signal to GPIO pin number */
#define GPIO_TO_PIN(bank, gpio) (32 * (bank) + (gpio))

/*!
 * \brief fops->ioctl ハンドラ
 *
 * \param file [IN] file構造体のポインタ
 * \param command [IN] IOCTLコマンド
 * \param arg [IN] 引数
 *
 * \return 0 : 成功
 */
static long
gpio_ioctl(struct file *file, unsigned int command, unsigned long arg)
{
    int rc;

    switch (command) {
    case GPIO_CTRL_IOCTL_CMD_PARAM:
    {
        unsigned long port;
        gpio_ctrl_ioctl_param p;

        rc = copy_from_user(&p, (void*)arg, sizeof(p));
        if (rc != 0) {
            printk("%s %d %s %s copy_from_user() failed\n",
                   __FILE__, __LINE__, __FUNCTION__, GPIO_CTRL_DEV_NAME);
            return (-EACCES);
        }

        switch (p.port) {
        case GPIO_CTRL_PORT_PHY_N_RST:
            port = GPIO_TO_PIN(1, 22);
            break;
        case GPIO_CTRL_PORT_NCPU_ADT_MUTE:
            port = GPIO_TO_PIN(1, 23);
            break;
        default:
            printk("%s %d %s %s Invalid Ioctl Command\n",
                   __FILE__, __LINE__, __FUNCTION__, GPIO_CTRL_DEV_NAME);
            return (-EINVAL);
        }
        if (p.direction) {
            gpio_direction_output(port, p.value);
        }
        else {
            gpio_direction_input(port);
            p.value = gpio_get_value(port);
            copy_to_user((void*)arg, &p, sizeof(p));
        }
        break;
    }
    default:
        printk("%s %d %s %s Invalid Ioctl Command\n",
               __FILE__, __LINE__, __FUNCTION__, GPIO_CTRL_DEV_NAME);
        return (-EINVAL);
    }

    return (0);
}

#define MAX_REG_MINOR (1)

static struct cdev  cdev;
static dev_t        dev;
static struct class *class;

/*!
 * \brief ドライバのロード処理
 *
 * \return 0 : 成功
 */
static int
gpio_ctrl_init(void)
{
    dev_t devno;
    int rc;

    /* メジャー番号の動的な取得 */
    rc = alloc_chrdev_region(&devno, 0, MAX_REG_MINOR, GPIO_CTRL_DEV_NAME);
    if (rc != 0) {
        printk("%s %d %s alloc_chrdev_region(%s) failed\n",
               __FILE__, __LINE__, __FUNCTION__, GPIO_CTRL_DEV_NAME);
        return (-ENOMEM);
    }

    /* キャラクタ・デバイスのデバイス番号の登録 */
    dev = MKDEV(MAJOR(devno), 0);

    cdev_init(&cdev, &gpio_ctrl_fops);
    rc = cdev_add(&cdev, dev, MAX_REG_MINOR);
    if (rc != 0) {
        printk("%s %d %s %s cdev_add() = %d\n",
               __FILE__, __LINE__, __FUNCTION__, GPIO_CTRL_DEV_NAME, rc);
        unregister_chrdev_region(dev, MAX_REG_MINOR);
        return (-ENOMEM);
    }

    /* /dev/xxx のデバイスノードの作成 */
    class = class_create(THIS_MODULE, GPIO_CTRL_DEV_NAME);
    if (IS_ERR(class)) {
        printk("%s %d %s %s class_create() failed\n",
               __FILE__, __LINE__, __FUNCTION__, GPIO_CTRL_DEV_NAME);
        cdev_del(&cdev);
        unregister_chrdev_region(dev, MAX_REG_MINOR);
        return (PTR_ERR(class));
    }

    device_create(class, NULL, dev, NULL, GPIO_CTRL_DEV_NAME);

    return (0);
}

/*!
 * \brief ドライバのアンロード処理
 *
 * \return 0 : 成功
 */
static void
gpio_ctrl_exit(void)
{
    unregister_chrdev_region(dev, MAX_REG_MINOR);
    device_destroy(class, dev);
    class_destroy(class);
    cdev_del(&cdev);
}

module_init(gpio_ctrl_init);
module_exit(gpio_ctrl_exit);
