
#include "cmd.h"
#include <linux/uaccess.h>  /* for copy_from_user / copy_to_user */
#include <linux/delay.h>

static int cmd4PP(cmdCtx *ctx, loff_t offs, const char __user *buf, size_t bufsize);
static void cmdRDSR1(cmdCtx *ctx, unsigned char *st);
static void cmdWREN(cmdCtx *ctx);
static void cmd4SE(cmdCtx *ctx, loff_t offs);

#define LOG(msg) printk("%s %d %s %s\n", __FILE__, __LINE__, __FUNCTION__, msg)

int
cmdInit(cmdCtx *ctx, struct spi_device *spi_device)
{
    ctx->spi_device = spi_device;

    ctx->txbuf = kmalloc(CMD_TXBUF_SIZE, GFP_KERNEL|GFP_DMA);
    if (ctx->txbuf == NULL) {
        LOG("kmalloc() failed");
        return (-ENOMEM);
    }

    ctx->rxbuf = kmalloc(CMD_RXBUF_SIZE, GFP_KERNEL|GFP_DMA);
    if (ctx->rxbuf == NULL) {
        LOG("kmalloc() failed");
        kfree(ctx->txbuf);
        return (-ENOMEM);
    }

    return (0);
}

void
cmdExit(cmdCtx *ctx)
{
    kfree(ctx->txbuf);
    kfree(ctx->rxbuf);
}

int
cmdRead(cmdCtx *ctx, loff_t offs,  char __user *buf, size_t bufsize)
{
    long ret;
    int rc;

    if (bufsize > SPI_FLASH_READ_SIZE) {
        LOG("bufsize is too large");
        return (-EINVAL);
    }

    spi_message_init(&ctx->message);

    memset(ctx->xfer, 0, sizeof(ctx->xfer));

    /* 16MBを超える SPIフラッシュのアドレス指定は 4byte */
    ctx->txbuf[0] = CMD_4FAST_READ;
    ctx->txbuf[1] = (unsigned char)((offs >> 24) & 0xFF);
    ctx->txbuf[2] = (unsigned char)((offs >> 16) & 0xFF);
    ctx->txbuf[3] = (unsigned char)((offs >>  8) & 0xFF);
    ctx->txbuf[4] = (unsigned char)((offs) & 0xFF);
    ctx->txbuf[5] = 0; /* Dummy */

    ctx->xfer[0].tx_buf         = ctx->txbuf;
    ctx->xfer[0].rx_buf         = NULL;
    ctx->xfer[0].len            = 6;
    ctx->xfer[0].bits_per_word  = 0;
    ctx->xfer[0].speed_hz       = 48000000;

    spi_message_add_tail(&ctx->xfer[0], &ctx->message);

    /* !!! Turbo Mode & DMA使用時に末尾2ワード目が化ける問題の回避策で2ワード余計に読み込む必要がある */
    ctx->xfer[1].tx_buf         = NULL;
    ctx->xfer[1].rx_buf         = ctx->rxbuf;
    ctx->xfer[1].len            = roundup(bufsize, 4) + 8 /* !!! */;
    ctx->xfer[1].bits_per_word  = 32;
    ctx->xfer[1].speed_hz       = 48000000;

    spi_message_add_tail(&ctx->xfer[1], &ctx->message);

    rc = spi_sync(ctx->spi_device, &ctx->message);
    if (rc != 0) {
        LOG("spi_sync() failed");
        return (rc);
    }

    ret = copy_to_user(buf, &ctx->rxbuf[0], bufsize);
    if (ret != 0) {
        LOG("copy_to_user() failed");
        return (ret);
    }

    return (0);
}

int
cmdWrite(cmdCtx *ctx, loff_t offs,  char __user *buf, size_t bufsize)
{
    const char *sp, *bp;
    int i, ret;
    unsigned char st;
    size_t write_size;

    sp = buf;
    bp = sp + bufsize;

    while ((bp - sp) > 0) {
        cmdWREN(ctx);

        for (i = 0; i < CMD_RETRY_MAX; i++) {
            cmdRDSR1(ctx, &st);
            if ((st & CMD_SR_WEL) != 0) {
                break;
            }
            LOG("waiting..");
            udelay(CMD_DELAY);
        }
        if (i >= CMD_RETRY_MAX) {
            LOG("Time out Write Enable");
            return (-EIO);
        }

        write_size = min(CMD_4PP_DATA_MAX, bp - sp);

        ret = cmd4PP(ctx, offs, sp, write_size);
        if (ret != 0) {
            return (ret);
        }

        sp   += write_size;
        offs += write_size;

        for (i = 0; i < CMD_RETRY_MAX; i++) {
            cmdRDSR1(ctx, &st);
            if ((st & CMD_SR_WIP) == 0) {
                break;
            }
            udelay(CMD_DELAY);
        }
        if (i >= CMD_RETRY_MAX) {
            LOG("Time out finish writing");
            return (-EIO);
        }
    }

    return (0);
}

int
cmdErase(cmdCtx *ctx, loff_t offs)
{
    unsigned char st;
    int i;

    if ((offs % CMD_4SE_ERASE_SIZE) > 0) {
        LOG("Invalid Offset");
        return (-EINVAL);
    }
    
    cmdWREN(ctx);

    for (i = 0; i < CMD_RETRY_MAX; i++) {
        cmdRDSR1(ctx, &st);
        if ((st & CMD_SR_WEL) != 0) {
            break;
        }
        LOG("waiting..");
        udelay(CMD_DELAY);
    }
    if (i >= CMD_RETRY_MAX) {
        LOG("Time out Write Enable");
        return (-EIO);
    }

    cmd4SE(ctx, offs);

    for (i = 0; i < CMD_RETRY_MAX; i++) {
        cmdRDSR1(ctx, &st);
        if ((st & CMD_SR_WIP) == 0) {
            break;
        }
        msleep(0);
    }
    if (i >= CMD_RETRY_MAX) {
        LOG("Time out finish writing");
        return (-EIO);
    }

    return (0);
}

static int
cmd4PP(cmdCtx *ctx, loff_t offs, const char __user *buf, size_t bufsize)
{
    long ret;
    int rc, i, s;
    char *p;

    /* Turbo Modeはリードオンリーモードでしか動作しないため、Turbo Modeのデータ化けはノーケア */
    if (bufsize > CMD_4PP_DATA_MAX) {
        LOG("bufsize is too large");
        return (-EINVAL);
    }

    /* 16MBを超える SPIフラッシュのアドレス指定は 4byte */
    ctx->txbuf[0] = CMD_4PP;
    ctx->txbuf[1] = (unsigned char)((offs >> 24) & 0xFF);
    ctx->txbuf[2] = (unsigned char)((offs >> 16) & 0xFF);
    ctx->txbuf[3] = (unsigned char)((offs >>  8) & 0xFF);
    ctx->txbuf[4] = (unsigned char)((offs) & 0xFF);

    ret = copy_from_user(&ctx->txbuf[CMD_4PP_CMD_SIZE], buf, bufsize);
    if (ret != 0) {
        LOG("copy_from_user() failed");
        return (ret);
    }

    /* 128 byte以上のxferを使用するとDMAモードになるため、分割して送信する */
    spi_message_init(&ctx->message);

    memset(ctx->xfer, 0, sizeof(ctx->xfer));

    p = &ctx->txbuf[0];
    s = CMD_4PP_CMD_SIZE + bufsize;

    for (i = 0; i < ARRAY_SIZE(ctx->xfer); i++) {
        ctx->xfer[i].tx_buf = (void*)p;
        ctx->xfer[i].len    = min(128, s);

        s -= ctx->xfer[i].len;
        p += ctx->xfer[i].len;

        spi_message_add_tail(&ctx->xfer[i], &ctx->message);

        if (s == 0) {
            break;
        }
    }

    rc = spi_sync(ctx->spi_device, &ctx->message);
    if (rc != 0) {
        LOG("spi_sync() failed");
        return (rc);
    }

    return (0);
}

static void
cmdRDSR1(cmdCtx *ctx, unsigned char *st)
{
    ctx->txbuf[0] = CMD_RDSR1;
    ctx->rxbuf[0] = 0;

    spi_write_then_read(ctx->spi_device, 
                        ctx->txbuf, 1,
                        ctx->rxbuf, 1);
    *st = ctx->rxbuf[0];
    return;
}

static void
cmdWREN(cmdCtx *ctx)
{
    ctx->txbuf[0] = CMD_WREN;

    spi_write_then_read(ctx->spi_device, 
                        ctx->txbuf, 1,
                        NULL, 0);

    return;
}

static void
cmd4SE(cmdCtx *ctx, loff_t offs)
{
    /* 16MBを超える SPIフラッシュのアドレス指定は 4byte */
    ctx->txbuf[0] = CMD_4SE;
    ctx->txbuf[1] = (unsigned char)((offs >> 24) & 0xFF);
    ctx->txbuf[2] = (unsigned char)((offs >> 16) & 0xFF);
    ctx->txbuf[3] = (unsigned char)((offs >>  8) & 0xFF);
    ctx->txbuf[4] = (unsigned char)((offs) & 0xFF);

    spi_write_then_read(ctx->spi_device, 
                        ctx->txbuf, 5,
                        NULL, 0);
    return;
}
