/*!
 * \file cmd.h
 * \brief SPI フラッシュコマンド
 * \author Hisashi Iida
 *
 * \note
 * Copyright(C) 2014 YAMAHA CORPORATION<br>
 * All rights reserved.<br>
 * $Id:$
 */
#ifndef __CMD_H__
#define __CMD_H__

#include "spi_flash_if.h"

#include <linux/completion.h>
#include <linux/spi/spi.h>

/*
 * R:\ＡＶレシーバ\14開発モデル\Vx79_barrack\技術資料\電気\NET\部品資料\SPI Flash\Spansion
 * S25FL128S_256S_00.pdf
 * 
 * 10.1 Command Set Summary
 * 
 *  4FAST_READ      Read Fast (4-byte Address)              0C
 *  4READ           Read (4-byte Address)                   13
 *  4DOR            Read Dual Out (4-byte Address)          3C
 *  4QOR            Read Quad Out (4-byte Address)          6C
 *  4DIOR           Dual I/O Read (4-byte Address)          BC
 *  4QIOR           Quad I/O Read (4-byte Address)          EC
 *  4DDRFR          Read DDR Fast (4-byte Address)          0E
 *  4DDRDIOR        DDR Dual I/O Read (4-byte Address)      BE
 *  4DDRQIOR        DDR Quad I/O Read (4-byte Address)      EE
 *  4PP             Page Program (4-byte Address)           12
 *  4QPP            Quad Page Program (4-byte Address)      34
 *  4P4E            Parameter 4-kB Erase (4-byte Address)   21
 *  4SE             Erase 64/256 kB (4-byte Address)        DC
 *
 * Table 10.2 S25FL128S and S25FL256S Command Set (sorted by function) (Sheet 1 of 2)
 *
 *  RDID            Read ID (JEDEC Manufacturer ID and JEDEC CFI)   9F
 *  RDSR1           Read Status Register-1                          05
 *  WREN            Write Enable                                    06
 */

#define CMD_4FAST_READ              (0x0C)
#define CMD_WREN                    (0x06)
#define CMD_RDID                    (0x9F)
#define CMD_RDSR1                   (0x05)
#define CMD_4PP                     (0x02)
#define CMD_4SE                     (0xDC)
#define CMD_4PP_CMD_SIZE            (5)
#define CMD_4PP_DATA_MAX            (256) /* spansion は 512が動作しない */
#define CMD_4SE_ERASE_SIZE          (SPI_FLASH_PASE_SIZE)

#define CMD_SR_WIP                  (1 << 0)
#define CMD_SR_WEL                  (1 << 1)

#define CMD_TXBUF_SIZE              (CMD_4PP_CMD_SIZE + CMD_4PP_DATA_MAX)
#define CMD_RXBUF_SIZE              (SPI_FLASH_READ_SIZE + 8 /* !!! */)
/* !!! Turbo Mode & DMA使用時に末尾2ワード目が化ける問題の回避策で2ワード余計に読み込む必要がある */

#define CMD_RETRY_MAX               (100000)
#define CMD_DELAY                   (100)

#define SPI_OMAP2_MCSPI_DMA_THRES   (128)

#define CMD_XFER_NUM ((CMD_TXBUF_SIZE + SPI_OMAP2_MCSPI_DMA_THRES - 1) / SPI_OMAP2_MCSPI_DMA_THRES)

typedef struct __YAMAHA_cmdCtx_struct {
    char                *txbuf;
    char                *rxbuf;

    struct spi_device   *spi_device;

    struct spi_message  message;
    struct spi_transfer xfer[CMD_XFER_NUM];
} cmdCtx;

extern int  cmdInit(cmdCtx *ctx, struct spi_device *spi_device);
extern void cmdExit(cmdCtx *ctx);
extern int  cmdRead(cmdCtx *ctx, loff_t offs,  char __user *buf, size_t bufsize);
extern int  cmdWrite(cmdCtx *ctx, loff_t offs,  char __user *buf, size_t bufsize);
extern int  cmdErase(cmdCtx *ctx, loff_t offs);

#endif
