/*!
 * \file spi_flash_if.c
 * \brief SPI FLASH ドライバ
 * \author Hisashi IIDA
 *
 * \note
 * Copyright(C) 2014 Yamaha Corporation<br>
 * All rights reserved.<br>
 * $Id:$
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/uaccess.h>  /* for copy_from_user / copy_to_user */

#include <linux/fs.h>       /* for struct file_operations */
#include <linux/cdev.h>
#include <linux/slab.h>
#include <linux/spi/spi.h>
#include <linux/sched.h>
#include <linux/delay.h>

#include "spi_flash_if.h"
#include "cmd.h"

#define MAX_REG_MINOR (1)

/* class_create / device_create を使うためには下記の宣言が必要 */
MODULE_LICENSE("GPL"); 

#define LOG(msg) printk("%s %d %s %s\n", __FILE__, __LINE__, __FUNCTION__, msg)

static int spi_flash_init(void);
static void spi_flash_exit(void);
static int spi_flash_probe(struct spi_device *spi_device);
static int spi_flash_remove(struct spi_device *spi);
static int spi_flash_open(struct inode *, struct file *);
static int spi_flash_release(struct inode *, struct file *);
static long spi_flash_ioctl(struct file *file, unsigned int type, unsigned long a);

/* ファイルオペレーション構造体 
 * include/linux/fs.h
 */
static const struct file_operations spi_driver_fops = {
    .owner              = THIS_MODULE,
    .open               = spi_flash_open,
    .unlocked_ioctl     = spi_flash_ioctl,
    .release            = spi_flash_release,
};

/* ドライバコンテキスト */
typedef struct __YAMAHA_spiFlashCtx_struct {
    struct cdev                     cdev;
    dev_t                           dev;
    struct class                    *class;
    const struct file_operations    *fops;
    struct spi_device               *spi_device;
    struct spi_driver               *spi_driver;
    const char                      *name;
    bool                            opened;

    cmdCtx                          cmdCtx;
} spiFlashCtx;

static spiFlashCtx __spiFlashCtx;

/* SPIドライバ登録情報 */
static struct spi_driver spi_flash_driver = {
    .driver = {
        .name   = SPI_FLASH_DEVICE_NAME,
        .bus    = &spi_bus_type,
        .owner  = THIS_MODULE,
    },
    .probe      = spi_flash_probe,
    .remove     = spi_flash_remove,
};

module_init(spi_flash_init);
module_exit(spi_flash_exit);

/*!
 * \brief ドライバのロード処理
 *
 * \return 0 : 成功
 */
static int
spi_flash_init(void)
{
    spiFlashCtx *ctx = &__spiFlashCtx;
    int rc;

LOG("");

    ctx->name = SPI_FLASH_DEVICE_NAME;
    ctx->fops = &spi_driver_fops;
    ctx->spi_device = NULL;
    ctx->spi_driver = &spi_flash_driver;

    /* SPIドライバの登録 */
    rc = spi_register_driver(ctx->spi_driver);
    if (rc != 0) {
        LOG("spi_register_driver() failed");
        return (rc);
    }

    return (0);
}

/*!
 * \brief ドライバのアンロード処理
 *
 * \return 0 : 成功
 */
static void
spi_flash_exit(void)
{
    spiFlashCtx *ctx = &__spiFlashCtx;

LOG("");
   spi_unregister_driver(ctx->spi_driver);
    return;
}

static int
spi_flash_probe(struct spi_device *spi_device)
{
    spiFlashCtx *ctx = &__spiFlashCtx;
    dev_t devno;
    int rc;
    struct device *device;

LOG("");
    ctx->spi_device = spi_device;

    /* メジャー番号の動的な取得 */
    rc = alloc_chrdev_region(&devno, 0, MAX_REG_MINOR, ctx->name);
    if (rc != 0) {
        LOG("alloc_chrdev_region() failed");
        return (-ENOMEM);
    }

    /* キャラクタ・デバイスのデバイス番号の登録 */
    ctx->dev = MKDEV(MAJOR(devno), 0);

    cdev_init(&ctx->cdev, ctx->fops);
    rc = cdev_add(&ctx->cdev, ctx->dev, MAX_REG_MINOR);
    if (rc != 0) {
        LOG("cdev_add() failed");
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        return (-ENOMEM);
    }

    /* /dev/xxx のデバイスノードの作成 */
    ctx->class = class_create(THIS_MODULE, ctx->name);
    if (IS_ERR(ctx->class)) {
        LOG("class_create() failed");
        cdev_del(&ctx->cdev);
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        return (-ENOMEM);
    }

    device = device_create(ctx->class, NULL, ctx->dev, NULL, ctx->name);
    if (IS_ERR(device)) {
        LOG("device_create() failed");
        class_destroy(ctx->class);
        cdev_del(&ctx->cdev);
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        return (-ENOMEM);
    }

    rc = cmdInit(&ctx->cmdCtx, ctx->spi_device);
    if (rc != 0) {
        LOG("cmdInit() failed");
        device_destroy(ctx->class, ctx->dev);
        device_destroy(ctx->class, ctx->dev);
        class_destroy(ctx->class);
        cdev_del(&ctx->cdev);
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        return (-ENOMEM);
    }
    ctx->opened = false;

    return (0);
}


static int
spi_flash_remove(struct spi_device *spi)
{
    spiFlashCtx *ctx = &__spiFlashCtx;

LOG("");

    (void)spi;

    cmdExit(&ctx->cmdCtx);
    ctx->spi_device = NULL;

    unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
    device_destroy(ctx->class, ctx->dev);
    class_destroy(ctx->class);
    cdev_del(&ctx->cdev);

    return (0);
}

static int
spi_flash_open(struct inode *inode, struct file *file)
{
    spiFlashCtx *ctx = &__spiFlashCtx;

    if (ctx->opened) {
        LOG("already open");
        return (-EBUSY);
    }
    ctx->opened = true;

    return (0);
}

static int
spi_flash_release(struct inode *inode, struct file *file)
{
    spiFlashCtx *ctx = &__spiFlashCtx;

    if (ctx->opened) {
        ctx->opened = false;
    }

    return (0);
}

static long spi_flash_ioctl(struct file *file, unsigned int type, unsigned long a)
{
    spiFlashCtx *ctx = &__spiFlashCtx;
    int ret;

    switch (type) {
    case SPI_FLASH_IOCTL_CMD_FAST_READ:
    {
        spiFlashFastReadParam p;

        ret = copy_from_user(&p, (void *)a, sizeof(p));
        if (ret != 0) {
            LOG("copy_to_user() failed");
            return (-EINVAL);
        }
        ret = cmdRead(&ctx->cmdCtx, p.addr, p.buf, p.size);
        if (ret != 0) {
            LOG("cmdRead() failed");
            return (ret);
        }
        break;
    }
    case SPI_FLASH_IOCTL_CMD_ERASE:
    {
        spiFlashEraseParam p;

        ret = copy_from_user(&p, (void *)a, sizeof(p));
        if (ret != 0) {
            LOG("copy_from_user() failed");
            return (-EINVAL);
        }
        ret = cmdErase(&ctx->cmdCtx, p.addr);
        if (ret != 0) {
            LOG("cmdErase() failed");
            return (ret);
        }
        break;
    }
    case SPI_FLASH_IOCTL_CMD_WRITE:
    {
        spiFlashWriteParam p;

        ret = copy_from_user(&p, (void *)a, sizeof(p));
        if (ret != 0) {
            LOG("copy_from_user() failed");
            return (-EINVAL);
        }
        ret = cmdWrite(&ctx->cmdCtx, p.addr, p.buf, p.size);
        if (ret != 0) {
            LOG("cmdWrite() failed");
            return (ret);
        }
        
        break;
    }
    default:
        LOG("invalid type");
        return (-EINVAL);
    }

    return (0);
}
