/*!
 * \file usb_err_if.c
 * \brief USBエラー情報取得
 * \author Hisashi IIDA
 *
 * \note
 * Copyright(C) 2014 Yamaha Corporation<br>
 * All rights reserved.<br>
 * $Id:$
 */

#define DEBUG_LOG                   (0)     /*!< デバッグ時は 1 にする */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/uaccess.h>  /* for copy_from_user / copy_to_user */

#include <linux/fs.h>       /* for struct file_operations */
#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/slab.h>
#include <linux/usb.h>

#include "usb_err_if.h"

#define MAX_REG_MINOR       (1)


/* 非対応デバイス vendor / product ID
   認定試験では，非対応デバイスとして以下を用いる。
    VID = 0x1A0A, かつ，PID = 0x0201または0x0202
*/
#define UNSP_VENDOR_ID      (0x1A0A)
#define UNSP_PRODUCT_ID1    (0x0201)
#define UNSP_PRODUCT_ID2    (0x0202)

/* PET(試験機) vendor / product ID
    VID = 0x1A0A, PID = 0x0200
*/
#define PET_VENDOR_ID       (0x1A0A)
#define PET_PRODUCT_ID      (0x0200)

/* コンテキスト情報 */
#define DEVICE_NAME     "usb_err"

typedef struct {
    const char      *name;
    struct cdev     cdev;
    dev_t           dev;
    struct class    *class;

    struct usb_device       *udev;
    struct usb_interface    *interface;
    struct usb_class_driver *dev_class;

    wait_queue_head_t   sleep;

    int     intf_unsp_id;
    int     intf_pet_id;

    bool    err_unsupport_device;
    bool    err_hub_support;
    bool    err_device_no_response;
    bool    err_babble;
    unsigned long   err_babble_cnt;
} ctx_drv;
static ctx_drv  __ctx;

/* ファイルオペレーション構造体 */
static int usb_err_open(struct inode *, struct file *);
static long usb_err_ioctl(struct file *, unsigned int, unsigned long);
static unsigned int usb_err_poll(struct file *, poll_table *);
static int usb_err_release(struct inode *, struct file *);

static const struct file_operations usb_err_fops = {
    .owner          = THIS_MODULE,
    .open           = usb_err_open,
    .unlocked_ioctl = usb_err_ioctl,
    .poll           = usb_err_poll,
    .release        = usb_err_release,
};

struct usb_err_dev {
    struct usb_device       *udev;
    struct usb_interface    *interface;
    struct usb_class_driver *dev_class;

    ctx_drv                 *ctx;
};


extern int musb_get_vbuserr_cont(void);
extern void hub_reg_callback_device_response_err(void (*)(void));
extern void hub_reg_callback_hub_support_err(void (*)(void));
extern void ti_reg_callback_babble_err(void (*)(void));


#if DEBUG_LOG
#define DEBUG_LOG_INFO      printk
#else   /* DEBUG_LOG */
#define DEBUG_LOG_INFO(...)
#endif  /* DEBUG_LOG */

#define DEBUG_LOG_INFO_CONFIGURATION(__inf)                                 \
    do {                                                                    \
        struct usb_device *__dev = usb_get_dev(interface_to_usbdev(__inf)); \
        int __num  = __dev->actconfig->desc.bNumInterfaces;                 \
        int __cnum = __dev->actconfig->desc.bConfigurationValue;            \
        struct usb_interface *__interface;                                  \
        int __i, __j;                                                       \
        DEBUG_LOG_INFO("  config %d(%d),if num=%d,cur(if=%d,alt=%d),len=%d:[%s] \n"\
            , __dev->actconfig->desc.iConfiguration                         \
            , __cnum                                                        \
            , __num                                                         \
            , __inf->cur_altsetting->desc.bInterfaceNumber                  \
            , __inf->cur_altsetting->desc.bAlternateSetting                 \
            , __dev->actconfig->desc.wTotalLength                           \
            , __dev->config[__cnum - 1].string                              \
        );                                                                  \
        for(__i = 0 ; __i < __num ; ++ __i){                                \
            __interface = __dev->config[__cnum - 1].interface[__i];         \
            if(__interface != NULL){                                        \
                struct usb_host_interface *__altset;                        \
                for(__j = 0 ; __j < __interface->num_altsetting ; ++ __j){  \
                    __altset = &(__interface->altsetting[__j]);             \
                    DEBUG_LOG_INFO("    if[%d]:typ=%d:alt=%d,enp=%d"        \
                        , __altset->desc.bInterfaceNumber                   \
                        , __altset->desc.bDescriptorType                    \
                        , __altset->desc.bAlternateSetting                  \
                        , __altset->desc.bNumEndpoints                      \
                    );                                                      \
                    if(__altset->endpoint != NULL){                         \
                        DEBUG_LOG_INFO("(%d)"                               \
                            , __altset->endpoint->desc.wMaxPacketSize       \
                        );                                                  \
                    }                                                       \
                    DEBUG_LOG_INFO(",cls=%d,sub=%d,ext=%d:[%s] \n"          \
                        , __altset->desc.bInterfaceClass                    \
                        , __altset->desc.bInterfaceSubClass                 \
                        , (unsigned int)__altset->extralen                  \
                        , __altset->string                                  \
                    );                                                      \
                }                                                           \
            }                                                               \
            else {                                                          \
                DEBUG_LOG_INFO("    if[%d]:NULL \n"                         \
                    , __interface->altsetting[0].desc.bInterfaceNumber      \
                );                                                          \
            }                                                               \
        }                                                                   \
    } while(0);


/*!
 * \brief ファイル open
 *
 * \return 0 : 成功
 */
static int
usb_err_open(struct inode *inode, struct file *file)
{
    ctx_drv *ctx;

    ctx = &__ctx;
    file->private_data = ctx;

    return 0;
}

/*!
 * \brief オーバーカレント回数取得
 *
 * \return 回数
 */
static int
ioctl_get_vbuserr_cont(void)
{
    return musb_get_vbuserr_cont();
}

/*!
 * \brief ファイル ioctl
 *
 * \return 0 : 成功
 */
static long
usb_err_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    ctx_drv *ctx;
    int ret;

    ctx = file->private_data;
    ret = 0;
    switch (cmd) {
    case USB_OVR_IOCTL_CMD_GET_VBUSERR_CONT:
        *((int *)arg) = ioctl_get_vbuserr_cont();
        break;

    case USB_OVR_IOCTL_CMD_GET_ERROR_EVENT:
        if (ctx->err_unsupport_device == true) {
            ctx->err_unsupport_device = false;
            *((USB_ERR_EVENT *)arg) = USB_ERR_EVENT_UNSUPPORT_DEVICE;
        }
        else if (ctx->err_hub_support == true) {
            ctx->err_hub_support = false;
            *((USB_ERR_EVENT *)arg) = USB_ERR_EVENT_HUB_NOT_SUPPORT;
        }
        else if (ctx->err_device_no_response == true) {
            ctx->err_device_no_response = false;
            *((USB_ERR_EVENT *)arg) = USB_ERR_EVENT_DEVICE_NO_RESPONSE;
        }
        else {
            *((USB_ERR_EVENT *)arg) = USB_ERR_EVENT_NONE;
        }
        break;
    case USB_OVR_IOCTL_CMD_GET_STS_UNSUPPORT_DEVICE:
        if (ctx->err_unsupport_device == true) {
            *((USB_ERR_EVENT *)arg) = USB_ERR_STS_TRUE;
        }
        else {
            *((USB_ERR_EVENT *)arg) = USB_ERR_STS_FALSE;
        }
        break;
    case USB_OVR_IOCTL_CMD_GET_STS_HUB_NOT_SUPPORT:
        if (ctx->err_hub_support == true) {
            *((USB_ERR_EVENT *)arg) = USB_ERR_STS_TRUE;
        }
        else {
            *((USB_ERR_EVENT *)arg) = USB_ERR_STS_FALSE;
        }
        break;
    case USB_OVR_IOCTL_CMD_GET_STS_DEVICE_NO_RESPONSE:
        if (ctx->err_device_no_response == true) {
            *((USB_ERR_EVENT *)arg) = USB_ERR_STS_TRUE;
        }
        else {
            *((USB_ERR_EVENT *)arg) = USB_ERR_STS_FALSE;
        }
        break;
    default:
        printk("[ERR]%s:%d::ioctl(0x%X) \n", __FUNCTION__, __LINE__, cmd);
        break;
    }

    return ret;
}

/*!
 * \brief ファイル poll
 *
 * \return 
 */
static unsigned int
usb_err_poll(struct file *file, poll_table *wait)
{
    ctx_drv *ctx;

    ctx = file->private_data;

    if (ctx->err_unsupport_device == true) {
        return POLLIN;
    }

    if (ctx->err_hub_support == true) {
        return POLLIN;
    }

    if (ctx->err_device_no_response == true) {
        return POLLIN;
    }
    poll_wait(file, &ctx->sleep, wait);

    return 0;
}

/*!
 * \brief ファイル release
 *
 * \return 0 : 成功
 */
static int
usb_err_release(struct inode *inode, struct file *file)
{
    ctx_drv *ctx;

    ctx = file->private_data;

    return 0;
}


/*!
 * \brief エニュメレーション失敗
 */
static void
callback_device_response_err(void)
{
    ctx_drv *ctx;

    ctx = &__ctx;
    ctx->err_device_no_response = true;
    wake_up(&ctx->sleep);
}

/*!
 * \brief ハブ未対応
 */
static void
callback_hub_support_err(void)
{
    ctx_drv *ctx;

    ctx = &__ctx;
    ctx->err_hub_support = true;
    wake_up(&ctx->sleep);
}

/*!
 * \brief バブルエラー対応
 */
static void
callback_babble_err(void)
{
    ctx_drv *ctx;

    ctx = &__ctx;
    ctx->err_babble = true;

    ++ ctx->err_babble_cnt;
    /* ※ここで表示(printk)すると、エラーが連続して発生してしまう */

//    wake_up(&ctx->sleep);
}


////////////////////////////////////////////////////////////////////////////////
static const struct file_operations pet_fops = {
    .owner = THIS_MODULE,
};

#define NUM_PET_CLASS       3
static char pet_dev_name[NUM_PET_CLASS][20];
static struct usb_class_driver pet_dev_class[NUM_PET_CLASS];

static struct usb_device_id usb_pet_id_table[] = {
    {   .match_flags  = USB_DEVICE_ID_MATCH_VENDOR | USB_DEVICE_ID_MATCH_PRODUCT,
        .idVendor     = PET_VENDOR_ID,
        .idProduct    = PET_PRODUCT_ID,
    },
    {},
};
MODULE_DEVICE_TABLE (usb, usb_pet_id_table);

/*!
 * \brief 接続
 *
 * \param intf [IN] 
 * \param id   [IN] 
 *
 * \retval 正常：0、異常：0以外
 */
static int
usb_pet_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
    struct usb_device *udev;
    struct usb_err_dev *dev;
    ctx_drv *ctx;
    int ret;

    ctx = &__ctx;

    DEBUG_LOG_INFO_CONFIGURATION(intf);

    /* インターフェース追加判定 */
    if (NUM_PET_CLASS <= ctx->intf_pet_id) {
        err("interface over error");
        return -ENOMEM;
    }

    dev = kzalloc(sizeof(struct usb_err_dev), GFP_KERNEL);
    if(dev == NULL){
        err("Out of memory");
        return -ENOMEM;
    }

    udev = usb_get_dev(interface_to_usbdev(intf));

    dev->udev = udev;
    dev->interface = intf;
    dev->dev_class = &pet_dev_class[ctx->intf_pet_id];
    dev->ctx       = ctx;

    /* インターフェース登録 */
    usb_set_intfdata(intf, dev);
    ret = usb_register_dev(intf, dev->dev_class);
    if(ret){
        kfree(dev);
        err("probe error %d", ret);
        return ret;
    }

    ++ ctx->intf_pet_id;

    dev_info(&intf->dev, "PET device probe (interface id=%d)\n"
        , ctx->intf_pet_id
    );

    return 0;
}

/*!
 * \brief 切断
 *
 * \param intf [IN] 
 *
 * \retval 正常：0、異常：0以外
 */
static void
usb_pet_disconnect(struct usb_interface *intf)
{
    ctx_drv *ctx;
    struct usb_err_dev *dev;

    dev = usb_get_intfdata(intf);
    if (dev == NULL) {
        return;
    }

    ctx = dev->ctx;
    ctx->err_babble = false;

    if (1 <= ctx->intf_pet_id) {
        -- ctx->intf_pet_id;
    }
    else {
        err("interface counter error %d", ctx->intf_pet_id);
        ctx->intf_pet_id = 0;
    }

    usb_set_intfdata(intf, NULL);
    usb_deregister_dev(intf, dev->dev_class);
    dev->interface = NULL;
    kfree(dev);

    dev_info(&intf->dev, "PET device disconnect (interface id=%d)\n"
        , ctx->intf_pet_id
    );
}

struct usb_driver usb_pet_driver = {
    .name       = "usb_pet",
    .probe      = usb_pet_probe,
    .disconnect = usb_pet_disconnect,
    .id_table   = usb_pet_id_table,
};


////////////////////////////////////////////////////////////////////////////////
static const struct file_operations unsp_fops = {
    .owner      = THIS_MODULE,
};

#define NUM_UNSP_CLASS      5
static char unsp_dev_name[NUM_UNSP_CLASS][20];
static struct usb_class_driver unsp_dev_class[NUM_UNSP_CLASS];

static struct usb_device_id usb_unsp_id_table[] = {
//    /* PET用 */
//    {   .match_flags  = USB_DEVICE_ID_MATCH_VENDOR | USB_DEVICE_ID_MATCH_PRODUCT,
//        .idVendor     = UNSP_VENDOR_ID,
//        .idProduct    = UNSP_PRODUCT_ID1,
//    },
//    {   .match_flags  = USB_DEVICE_ID_MATCH_VENDOR | USB_DEVICE_ID_MATCH_PRODUCT,
//        .idVendor     = UNSP_VENDOR_ID,
//        .idProduct    = UNSP_PRODUCT_ID2,
//    },
    /* その他 */
//    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
//        .bDeviceClass       = USB_CLASS_PER_INTERFACE
//    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_AUDIO,
        .bInterfaceSubClass	= 1,
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_COMM
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_HID
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_PHYSICAL
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_STILL_IMAGE
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_PRINTER
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_MASS_STORAGE
    },
//ハブドライバで検出させるため、ここでは定義しない
//    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
//        .bInterfaceClass    = USB_CLASS_HUB
//    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_CDC_DATA
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_CSCID
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_CONTENT_SEC
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_VIDEO
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_WIRELESS_CONTROLLER
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_MISC
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_APP_SPEC
    },
    {   .match_flags        = USB_DEVICE_ID_MATCH_INT_CLASS,
        .bInterfaceClass    = USB_CLASS_VENDOR_SPEC
    },
    {},
};
MODULE_DEVICE_TABLE (usb, usb_unsp_id_table);

/*!
 * \brief 接続
 *
 * \param intf [IN] 
 * \param id   [IN] 
 *
 * \retval 正常：0、異常：0以外
 */
static int
usb_unsp_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
    struct usb_device *udev;
    struct usb_err_dev *dev;
    ctx_drv *ctx;
    int ret;
    int num;
    int cnum;
    struct usb_interface *interface;
    int i;
    int j;

    ctx = &__ctx;

    DEBUG_LOG_INFO_CONFIGURATION(intf);

    /* インターフェース追加判定 */
    if (NUM_UNSP_CLASS <= ctx->intf_unsp_id) {
        err("interface over error");
        return -ENOMEM;
    }

    /* 現在のコンフィグ中に mass storage が存在するかの判定 */
    udev = usb_get_dev(interface_to_usbdev(intf));
    num  = udev->actconfig->desc.bNumInterfaces;
    cnum = udev->actconfig->desc.bConfigurationValue;
    for(i = 0 ; i < num ; ++ i){
        interface = udev->config[cnum - 1].interface[i];
        if(interface != NULL){
            struct usb_host_interface *altset;
            for(j = 0 ; j < interface->num_altsetting ; ++ j){
                altset = &(interface->altsetting[j]);
                if (altset != NULL) {
                    if (altset->desc.bInterfaceClass == USB_CLASS_MASS_STORAGE) {
                        return -ENOMEM;
                    }
                }
            }
        }
    }

    dev = kzalloc(sizeof(struct usb_err_dev), GFP_KERNEL);
    if(dev == NULL){
        err("Out of memory");
        return -ENOMEM;
    }

    udev = usb_get_dev(interface_to_usbdev(intf));

    dev->udev = udev;
    dev->interface = intf;
    dev->dev_class = &unsp_dev_class[ctx->intf_unsp_id];
    dev->ctx       = ctx;

    /* インターフェース登録 */
    usb_set_intfdata(intf, dev);
    ret = usb_register_dev(intf, dev->dev_class);
    if(ret){
        kfree(dev);
        err("probe error %d", ret);
        return ret;
    }

    ++ ctx->intf_unsp_id;

    if (ctx->intf_unsp_id == 1) {
        /* 最初のみ通知 */
        ctx->err_unsupport_device = true;
        wake_up(&ctx->sleep);
    }

    dev_info(&intf->dev, "unsupported device probe (interface id=%d)\n"
        , ctx->intf_unsp_id
    );

    return 0;
}

/*!
 * \brief 切断
 *
 * \param intf [IN] 
 *
 * \retval 正常：0、異常：0以外
 */
static void
usb_unsp_disconnect(struct usb_interface *intf)
{
    ctx_drv *ctx;
    struct usb_err_dev *dev;

    dev = usb_get_intfdata(intf);
    if (dev == NULL) {
        return;
    }

    ctx = dev->ctx;
    ctx->err_unsupport_device = false;
    ctx->err_babble = false;
    ctx->err_babble_cnt = 0;

    if (1 <= ctx->intf_unsp_id) {
        -- ctx->intf_unsp_id;
    }
    else {
        err("interface counter error %d", ctx->intf_unsp_id);
        ctx->intf_unsp_id = 0;
    }

    usb_set_intfdata(intf, NULL);
    usb_deregister_dev(intf, dev->dev_class);
    dev->interface = NULL;
    kfree(dev);

    dev_info(&intf->dev, "unsupported device disconnect (interface id=%d)\n"
        , ctx->intf_unsp_id
    );
}

struct usb_driver usb_unsp_driver = {
    .name       = "usb_unsp",
    .probe      = usb_unsp_probe,
    .disconnect = usb_unsp_disconnect,
    .id_table   = usb_unsp_id_table,
};


////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief ドライバのロード処理
 *
 * \return 0 : 成功
 */
static int
usb_err_init(void)
{
    ctx_drv *ctx;
    dev_t devno;
    struct usb_class_driver *cl;
    int rc;
    int i;

    ctx = &__ctx;

    ctx->intf_unsp_id = 0;
    for (i = 0 ; i < NUM_UNSP_CLASS ; ++ i) {
        cl = &unsp_dev_class[i];
        sprintf((char *)&unsp_dev_name[i], "usb_unsp%d", i + 1);
        cl->name       = (char *)&unsp_dev_name[i];
        cl->fops       = &unsp_fops;
        cl->minor_base = MAX_REG_MINOR;
    }

    ctx->intf_pet_id = 0;
    for (i = 0 ; i < NUM_PET_CLASS ; ++ i) {
        cl = &pet_dev_class[i];
        sprintf((char *)&pet_dev_name[i], "usb_pet%d", i + 1);
        cl->name       = (char *)&pet_dev_name[i];
        cl->fops       = &pet_fops;
        cl->minor_base = MAX_REG_MINOR;
    }

    /* コンテキスト初期化 */
    ctx->name = DEVICE_NAME;

    /* メジャー番号の動的な取得 */
    rc = alloc_chrdev_region(&devno, 0, MAX_REG_MINOR, ctx->name);
    if (rc != 0) {
        return (-ENOMEM);
    }

    /* キャラクタ・デバイスのデバイス番号の登録 */
    ctx->dev = MKDEV(MAJOR(devno), 0);

    cdev_init(&ctx->cdev, &usb_err_fops);
    rc = cdev_add(&ctx->cdev, ctx->dev, MAX_REG_MINOR);
    if (rc != 0) {
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        return (-ENOMEM);
    }

    /* /dev/xxx のデバイスノードの作成 */
    ctx->class = class_create(THIS_MODULE, ctx->name);
    if (IS_ERR(ctx->class)) {
        cdev_del(&ctx->cdev);
        unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
        return (PTR_ERR(ctx->class));
    }
    device_create(ctx->class, NULL, ctx->dev, NULL, ctx->name);

    /* poll用 */
    init_waitqueue_head(&ctx->sleep);

    /* エラー検出識別 */
    ctx->err_unsupport_device   = false;
    ctx->err_hub_support        = false;
    ctx->err_device_no_response = false;
    ctx->err_babble             = false;
    ctx->err_babble_cnt         = 0;

    /* PET用ドライバ */
    rc = usb_register(&usb_pet_driver);
    if (rc != 0) {
        err("usb_err(PET) register failed. Error number %d", rc);
    }

    /* 非対応デバイス用ドライバ */
    rc = usb_register(&usb_unsp_driver);
    if (rc != 0) {
        err("usb_err(unsupport) register failed. Error number %d", rc);
    }

    /* エラー検出登録 */
    hub_reg_callback_device_response_err(callback_device_response_err);
    hub_reg_callback_hub_support_err(callback_hub_support_err);
    ti_reg_callback_babble_err(callback_babble_err);

    return (0);
}

/*!
 * \brief ドライバのアンロード処理
 *
 * \return 0 : 成功
 */
static void
usb_err_exit(void)
{
    ctx_drv *ctx;

    ctx = &__ctx;

    /* エラー検出登録削除 */
    hub_reg_callback_device_response_err(NULL);
    hub_reg_callback_hub_support_err(NULL);
    ti_reg_callback_babble_err(NULL);

    /* PET用ドライバ */
    usb_deregister(&usb_pet_driver);
    /* 非対応デバイス用ドライバ */
    usb_deregister(&usb_unsp_driver);

    /* ドライバ削除 */
    device_destroy(ctx->class, ctx->dev);
    class_destroy(ctx->class);
    unregister_chrdev_region(ctx->dev, MAX_REG_MINOR);
    cdev_del(&ctx->cdev);

    return;
}

module_init(usb_err_init);
module_exit(usb_err_exit);
MODULE_LICENSE("GPL");
